import logging
import os
from enum import Enum

import boto3
import docker
from docker.errors import ImageNotFound

from quantit_snapshot.base.setting.settings import (
    SNAPSHOT_AWS_S3_KEY, SNAPSHOT_AWS_S3_SECRET_KEY, SNAPSHOT_PATH
)


class ImageEnv(Enum):
    sm_base = 'sm_base'

    @classmethod
    def help(cls):
        print_text = "Envrionment Now available :"
        n_line = '\n   - '
        env_name = [member.name for member in cls]
        for e in env_name:
            print_text += n_line + e

        example = """

Image download example) 
    from quantit_snapshot.base.setting.venv.docker import ImageEnv
    ImageEnv.pull_run_ecr(ImageEnv.sm_base)
        """
        print_text += example
        print(print_text)

    @staticmethod
    def get_clients():
        ecr_client = boto3.client(
            'ecr',
            region_name='ap-northeast-2',
            aws_access_key_id=SNAPSHOT_AWS_S3_KEY,
            aws_secret_access_key=SNAPSHOT_AWS_S3_SECRET_KEY,
        )
        docker_client = docker.from_env()
        return ecr_client, docker_client

    @staticmethod
    def pull_ecr(env_name, tag, **kwargs):
        ecr_client, docker_client = ImageEnv.get_clients()

        ecr_auth = ecr_client.get_authorization_token()
        ecr_auth = ecr_auth['authorizationData'][0]
        token = ecr_auth['authorizationToken']
        proxy_ep = ecr_auth['proxyEndpoint']
        end_point = proxy_ep.replace('http://', '').replace('https://', '')

        image_name = f'{end_point}/{env_name.value}:{tag}'

        try:
            img = docker_client.images.get(image_name)
            logging.warning("[WARNING] image already exist. Automatically use existing image.")
            return img
        except ImageNotFound:
            logging.warning("[INFO] Newly download ECR image.")

        docker_client.login(username='AWS', password=token, registry=end_point)
        img = docker_client.images.pull(image_name)  # 이미지가 있으면 다운로드 하지 않음.
        return img

    @staticmethod
    def run_container(
            short_id, snapshot_path, name, **kwargs
    ):
        _, docker_client = ImageEnv.get_clients()
        container = docker_client.containers.run(
            image=short_id.split(":")[1],
            name=name,
            volumes={
                snapshot_path: {"bind": "/locdisk/code/snapshot", "mode": "rw"}
            },
            environment={
                "AWS_ACCESS_KEY_ID": SNAPSHOT_AWS_S3_KEY,
                "AWS_SECRET_ACCESS_KEY": SNAPSHOT_AWS_S3_SECRET_KEY,
                "GITLAB_PERSONAL_ACCESS_TOKEN": os.environ["GITLAB_PERSONAL_ACCESS_TOKEN"]
            },
            detach=True,
            tty=True,
            command="/bin/bash"
        )
        return container

    @staticmethod
    def pull_run_ecr(env_name, container_name, tag="latest", snapshot_path=SNAPSHOT_PATH, **kwargs):
        pulled_img = ImageEnv.pull_ecr(env_name, tag, **kwargs)
        container = ImageEnv.run_container(
            pulled_img.short_id, snapshot_path, container_name
        )
